/* Package World Simulator
   Author: Jason Boire
   Significant modifications by: Jeff Heflin
*/
package pacworld;

import agent.*;
import java.util.Vector;
import java.util.Date;
import javax.swing.JOptionPane;


/** A simulator for the package delivery world environment. */
public class PackageWorld extends Environment{

  /** The number of milliseconds used by agents to process percepts and
      choose actions. */
  protected long thinkTime = 0;

  public static int DEFAULT_NUM_AGENTS = 4;

  /** The length (and width) of the world. The world must be a square. */
  public static int DEFAULT_WORLD_SIZE = 50;

  /** Number of packages. */
  public static int DEFAULT_NUM_PACKAGES = 20;

  /** Number of destinations. */
  public static int DEFAULT_NUM_DESTINATIONS = 3;
  public static int MAX_NUM_DESTINATIONS = 8;

  protected int numAgents = DEFAULT_NUM_AGENTS;
  protected int worldSize = DEFAULT_WORLD_SIZE;
  protected int numPackages = DEFAULT_NUM_PACKAGES;
  protected int numDestinations = DEFAULT_NUM_DESTINATIONS;


  public PackageWorld() {
    super();
  }


  /** Add a new agent to the environment. Note, the agents vector is
      actually a vector of PacAgentReps. */
  public void addAgent(Agent agent) {

    int id = agents.size();
    PacAgentRep arep = new PacAgentRep(id, (PacAgent)agent);
    agents.add(arep);
  }

  public Vector getAgents() {
    return agents;
  }

  public void resetAgents() {
    agents.removeAllElements();
  }
 

  /** The simulation is complete when all packages have been delivered or
    * nothing has happened for two cycles or no packages have been picked up
    * or dropped off for 5 * n cycles. */
  protected boolean isComplete() {
    PackageState s = (PackageState)state;

    if (s.getNumPackages() == 0 || s.getIdleCount() >= 2 * numAgents || s.getNonProductiveCount() >= 5 * numAgents * worldSize) {
      //s.printState();

      delivered = s.getOrigNumPackages() - s.getNumPackages();
      avgEffort = count /  s.getOrigNumPackages();
      msgLengthPerTurn = s.getTotalMessageLength() / count;
      thinkPerTurn = thinkTime / count;

      String str = "Simulation Complete.\nResults:";
      str += "\nPackages delivered: ";
      str += delivered + " / " + s.getOrigNumPackages();
      str += "\nTurns per package: " + avgEffort;
      str += "\nMessage length per turn: " + msgLengthPerTurn;
      str += "\nThink time per turn: " + thinkPerTurn;

    /*
      str += "\nTotal number of turns: ";
      str += count;
      str += "\nTotal Number of actions: ";    // not including IDLE
      str += s.getWorkCount();
      str += "\nMoves with packages: ";
      str += s.getPacCount();
      str += "\nTotal message count: ";
      str += s.getMessageCount();
      str += "\nTotal message length: ";
      str += s.getTotalMessageLength();
      str += "\nTotal think time (ms): ";
      str += thinkTime;
    */
      str += "\n\nPerformance Measure: ";
      str += getPerformanceMeasure();

      JOptionPane.showMessageDialog(null, str, "Final Results", JOptionPane.INFORMATION_MESSAGE);
      return true;
    }
    else
      return false;
  }

    /** Return the performance measure of the agent in the current 
      environment. Higher values are considered better. */
  protected int getPerformanceMeasure() {

    PackageState s = (PackageState)state;

    int perf = (int)((delivered / s.getOrigNumPackages() * 1000) - 
		     (avgEffort * 2) - (msgLengthPerTurn / 50) - thinkPerTurn);
    return perf;
  }

  
  /** Create a percept for an agent. This implements the see: S -> P
      function. */
  protected Percept getPercept(AgentRepresentation a) {
    
    PacPercept p;

    if (state instanceof PackageState) {
      PackageState s = (PackageState) state;
      PacAgentRep agt = (PacAgentRep) a;
      p = new PacPercept(s, agt);
      // System.out.println("Pecept: " + p.toString());
      return p;
    }
    else {
      System.out.println("ERROR - state is not a PackageState object.");
      return null;
    }
  }

  /** Execute an agent's action and update the environment's state. */
  protected void updateState(AgentRepresentation a, Action action) {

    super.updateState(a, action);
    count++;
  }

  public int getWorldSize() {
    return worldSize;
  }

  public int getNumAgents() {
    return numAgents;
  }

  public void setGUI(PacGUI gui) {
    pg = gui;
  }

    /*
  public void setInitState(PackageState state) {
    initState = state;
  }
    */

  /** Run the simulation starting from a given state. This consists of
      a sense-act loop for each agent. */
  public void start(State initState) {

    Percept p;
    Action action;
    PacAgentRep a;
    Date d1, d2;

    state = initState;
    state.display();

    while (!isComplete()) {
      for (int i=0; i < numAgents; i++) {
        a = (PacAgentRep)agents.elementAt(i);
        p = getPercept(a);
        d1 = new Date();
        a.see(p);
        action = a.selectAction();
        d2 = new Date();
        thinkTime += (d2.getTime() - d1.getTime());
	System.out.println("Agent " + a.getId() + ": Action = " + action);
	// System.out.println("Total think time= " + thinkTime + "ms");
        updateState(a, action);
        state.display();
      }
      System.out.println();
    }
  }


  /** Starts the program. */
  public static void main(String[] args) {

    PackageWorld world;
    PackageState initState;

    world = new PackageWorld();
    world.count = 0;

    if (args.length > 4) {
	System.out.println("Usage: java pacworld.PackageWorld [numAgents] [numPackages] [numDestinations] [worldSize]");
	System.exit(1);
    }
    if (args.length >= 1) {
      world.numAgents = Integer.parseInt(args[0]);
    }
    if (args.length >= 2) {
      world.numPackages = Integer.parseInt(args[1]);
    }
    if (args.length >= 3) {
      world.numDestinations = Integer.parseInt(args[2]);
      if (world.numDestinations > 8) {
	  System.out.println("A maximum of " + MAX_NUM_DESTINATIONS +
			     " destinations are allowed.");
	  System.exit(1);
      }	      
    }
    if (args.length == 4) {
      world.worldSize = Integer.parseInt(args[3]);
    }
      
    for (int i = 0; i < world.numAgents; i++) {
      PacAgent agent = new PacAgent(i);
      world.addAgent(agent);
    }

    initState = PackageState.getInitState(world.getAgents(), world.numPackages,
					  world.numDestinations, 
					  world.worldSize);
    PacGUI pg = new PacGUI(world);
    pg.setDefaultCloseOperation(javax.swing.JFrame.EXIT_ON_CLOSE);
    world.setGUI(pg);
    initState.setGUI(pg);

    world.start(initState);
    world.pg.dispose();
    System.exit(0);
  }

  protected PacGUI pg = null;
  protected PackageState initState = null;
  
  /** The number of turns (each agent's sense/act pair is a turn) */
  protected int count;
  
  /** The number of packages successfully delivered. */
  protected int delivered;

  protected float avgEffort;
  protected float msgLengthPerTurn;
  protected float thinkPerTurn;

}
